/*
  Hallo
  Example:

    Hallo
*/

#pragma once
#include <glm/glm.hpp>
#include <functional>
#include <memory>
#include <span>

#include "types.hpp"

enum TransportType
{
    TRANSPORT_TYPE_UDP
};

enum TransportState
{
    TRANSPORT_STATE_UNINITIALIZED,
    TRANSPORT_STATE_WAITING,
    TRANSPORT_STATE_CONNECTED,
    TRANSPORT_STATE_ERROR
};

class Transport
{
public:
    typedef std::shared_ptr<Transport> Ptr;

    typedef std::function<void(const glm::u32vec2& resolution)> OnSetup;
    typedef std::function<void(const glm::mat4& left_eye_projection, const glm::mat4& right_eye_projection, const glm::mat4& left_head_to_eye, const glm::mat4& right_head_to_eye)> OnProjectionChange;
    typedef std::function<void(TransformId transform, const glm::mat4& head_transform)> OnHeadTransform;
    typedef std::function<void(Controller controller, const glm::mat4& controller_transform)> OnControllerTransform;
    typedef std::function<void(Controller controller, ControllerState controller_state)> OnControllerEvent;
    typedef std::function<void(Controller controller, Button button, ButtonState button_state)> OnControllerButton;
    typedef std::function<void(Controller controller, const glm::vec2& thumbstick)> OnControllerThumbstick;
    typedef std::function<void(FrameNumber frame_number, FrameId frame_id, TransformId transform_id, double timestamp_transform_query, double timestamp_server_response, double timestamp_frame_decoded, double timestamp_command_construct, double timestamp_command_executed)> OnLatency;
    typedef std::function<void()> OnTransportError;

public:
    Transport() = default;
    virtual ~Transport() = default;

    virtual bool create(uint32_t port_number) = 0;
    virtual void destroy() = 0;
    virtual bool wait_connect() = 0;

    //NOTE: The following functions should be thread safe
    virtual bool send_setup_complete(RemoteStrategy remote_strategy, uint32_t max_frame_ids, float near_plane, float far_plane) = 0;
    virtual bool send_stage_transform(const glm::mat4& stage_transform) = 0;
    virtual bool send_frame_config_nal(FrameId frame_id, const std::span<uint8_t>& content) = 0;
    virtual bool send_frame_nal(FrameNumber frame_number, FrameId frame_id, TransformId transform_id, const std::span<uint8_t>& content) = 0;
    virtual bool send_frame_metadata(FrameNumber frame_number, const std::span<uint8_t>& content) = 0;

    //NOTE: The callbacks are only executed by the worker thread of the transport class.
    //NOTE: Set the callbacks before calling create and don't do it again after calling create.
    virtual void set_on_setup(OnSetup function) = 0;
    virtual void set_on_projection_change(OnProjectionChange function) = 0;
    virtual void set_on_head_transform(OnHeadTransform function) = 0;
    virtual void set_on_controller_transform(OnControllerTransform function) = 0;
    virtual void set_on_controller_event(OnControllerEvent function) = 0;
    virtual void set_on_controller_button(OnControllerButton function) = 0;
    virtual void set_on_controller_thumbstick(OnControllerThumbstick function) = 0;
    virtual void set_on_latency(OnLatency function) = 0;
    virtual void set_on_transport_error(OnTransportError function) = 0;

    //NOTE: The following functions should be thread safe
    virtual void set_max_send_queue_size(uint32_t size) = 0; //NOTE: The size is given in Bytes

    //NOTE: The following function should be thread safe
    virtual TransportState get_state() = 0;
    virtual double get_bitrate_send() = 0;          //NOTE: The bitrate is given in MBits/s
    virtual double get_bitrate_received() = 0;      //NOTE: The bitrate is given in MBits/s
    virtual uint32_t get_max_send_queue_size() = 0; //NOTE: The size is given in Bytes
    virtual uint32_t get_send_queue_size() = 0;     //NOTE: In Bytes
    virtual uint32_t get_receive_queue_size() = 0;  //NOTE: In Bytes
};

Transport::Ptr make_transport(TransportType transport_type);